package MMM::MirrorQueue;

use strict;
use warnings;
use URI;
use Sys::Hostname ();
use base qw(MMM::MirrorList);

sub new {
    my ( $class, $name, $mmm ) = @_;
    bless(
        {
            name         => $name,
            level_factor => 60,
            frequency    => 120,
            level        => 5,
            fetched      => 0,
            'log'        => $mmm ? sub { $mmm->log(@_) } : undef,
            hostname => ( $mmm ? $mmm->hostname() : undef )
              || Sys::Hostname::hostname(),
        },
        $class
    );
}

sub log {
    my ( $self, $level, $fmt, @args ) = @_;
    if ( $self->{log} ) {
        $self->{log}->( $level, $fmt, @args );
    }
    else {
        printf "$fmt\n", @args;
    }
}

sub level {
    my ($self) = @_;
    defined($self->{level}) ? $self->{level} : 5
}

sub frequency {
    my ($self) = @_;
    defined($self->{frequency}) ? $self->{frequency} : 5
}

sub hostname {
    my ($self) = @_;
    $self->{hostname}
}

sub reset {
    my ($self, %options) = @_;
    foreach (keys %options) {
        $self->{$_} = $options{$_};
    }
    $self->{sorted_mirror} = [];
    $self->{sorted}        = 0;
    $self->{fetched}       = 0;
}

sub fetched {
    return $_[0]->{fetched} || 0;
}

sub _sorted_wanted_level {
    my ($self) = @_;
    (sort { $b <=> $a } grep { $_ < $self->level } $self->list_levels),
    (sort { $a <=> $b } grep { $_ >= $self->level } $self->list_levels)
}

sub sort_list {
    my ($self) = @_;
    $self->{sorted} and return;
    foreach my $level ($self->_sorted_wanted_level()) {
        my @l;
        foreach my $mirror (
            sort {
                (abs($self->frequency - $a->frequency) + $a->frequency * $a->random) <=>
                (abs($self->frequency - $b->frequency) + $b->frequency * $b->random)
            } grep { $_->level == $level } $self->mirrors) {

            $mirror->host() eq $self->hostname and next;

            if ($self->{prefered_host}{$level} && $self->{prefered_host}{$level} eq $mirror->host) {
                unshift(@l, $mirror);
                next;
            }
            push(@l, $mirror);
        }

        push( @{ $self->{sorted_mirror} }, @l );
    }

    if ( @{ $self->{sorted_mirror} || [] } ) {
        if ( $self->{sorted_mirror}[0]->level >= $self->level ) {
            $self->log( 'WARNING',
                "cannot find any mirror for level %d for source %s",
                $self->level, $self->{sorted_mirror}[0]->source );
        }
    }
    else {
        $self->log( 'ERROR', "cannot find any mirror" );
    }
    $self->{fetched} = 0;
    $self->{sorted}  = 1;
}

sub get_next {
    my ($self) = @_;
    $self->sort_list();
    $self->{fetched}++;
    return $self->{current} = shift( @{ $self->{sorted_mirror} || [] } );
}

sub get_current {
    my ($self) = @_;
    $self->{current};
}

1;
