package MMM::Rsync;

use strict;
use warnings;
use Digest::MD5 qw(md5_hex);
use Config::IniFiles;
use MMM::MirrorQueue;
use IPC::Open3;
use IO::Select;

=head1 NAME

MMM::Rsync

=head1 SYNOPSIS

=head1 DESCRIPTION

An object to manage the rsync process and mirror
choice.

=head1 METHODS

=head2 new( $source, %options)

Create a new MMM::Rsync object

$source is the mirror source name

%options should provide:

statedir: the directory where status ini file are located
log: a code ref to handle message

=cut

sub new {
    my ( $class, $source, %options ) = @_;
    my $statusfile = "$options{statedir}/" . md5_hex( $options{dest} );
    bless(
        {
            mirrorlist => $options{mirrorlist} || MMM::MirrorQueue->new(),
            source     => $source,
            statusfile => $statusfile,
            status     => (
                Config::IniFiles->new(
                    -f $statusfile ? ( -file => $statusfile ) : ()
                  )
                  || Config::IniFiles->new(),
            ),
            %options,
        },
        $class
    );
}

=head2 log( $level, $fmt, @args)

This function is used to handle messages

=cut

sub log {
    my ( $self, $level, $fmt, @args ) = @_;
    if ( $self->{log} ) {
        $self->{log}->( $level, $fmt, @args );
    }
    else {
        printf "$fmt\n", @args;
    }
}

=head2 dest

Return the destination directory of this process

=cut

sub dest {
    return $_[0]->{dest};
}

=head2 source

Return the mirror source handle by the object

=cut

sub source {
    $_[0]->{source};
}

=head2 level

Return the level this queue should achieve

=cut

sub level {
    my ($self) = @_;
    my $level = $self->val( 'level', $self->{status}->val( 'job', 'level' ) );
    if ( !defined($level) ) {
        $level = $self->choose_my_level();
        $self->{status}->newval( 'job', 'level', $level );
    }
    $level;
}

=head2 frequency

Return the frequency value from config

=cut

sub frequency {
    my ($self) = @_;
    $self->val( 'period', 120 );
}

=head2 estimate_frequency

Return the estimate frequency after taking into account all delaying
configuration value

=cut

sub estimate_frequency {
    my ($self) = @_;
    my ($f) = sort { $b <=> $a }
      grep { defined($_) } (
        $self->frequency,
        $self->val('waitafter'),
        $self->val('waitaftersuccess'),
      );
    $f;
}

=head2 val( $var, $default)

Return the configuration value for $var. Return $default if parameter
is not set in the config

=cut

sub val {
    my ( $self, $var, $default ) = @_;
    $self->{config}->val( $self->{dest}, $var, $default );
}

=head2 mirrorlist

Return a ref to the MirrorQueue mirror collection used by this rsync queue

=cut

sub mirrorlist {
    $_[0]->{mirrorlist};
}

=head2 set_mirror_list( $mirrorlist)

Set the MirrorQueue mirrir collection used by this rsync queue

=cut

sub set_mirror_list {
    my ( $self, $mirrorlist ) = @_;
    $self->{mirrorlist} = $mirrorlist;
}

=head2 is_disable

Return true if the queue is disable

=cut

sub is_disable {
    my ($self) = @_;
    if ( $self->val('disable')
        && lc( $self->val('disable') ) =~ /yes|true|on|1/ )
    {
        return 1;
    }
    else {
        return;
    }
}

=head2 need_run

Return true if the queue have to run according schedule time

=cut

sub need_run {
    my ($self) = @_;
    $self->{status}->ReadConfig() if ( $self->{status}->GetFileName() );
    return $self->next_run_time <= scalar( time() );
}

=head2 is_running

Return true if the is already running

=cut

sub is_running {
    my ($self) = @_;
    my $oldpid = $self->{status}->val( 'job', 'pid' );
    if ( $oldpid && -d "/proc/$oldpid" ) {
        return 1;
    }
    0;
}

=head2 next_run_time

Return the next time estimate process time

=cut

sub next_run_time {
    my ($self) = @_;
    my @alltime = ( scalar( time() ) );

    if ( $self->last_start ) {
        push( @alltime, $self->last_start + ( $self->frequency * 60 ) );
    }

    if (   $self->val('waitafter')
        && $self->{status}->val( 'lastrun', 'end' ) )
    {
        push( @alltime,
            $self->{status}->val( 'lastrun', 'end' ) + $self->val('waitafter') *
              60 );
    }

    if (   $self->val('waitaftersuccess')
        && $self->{status}->val( 'lastsuccess', 'end' ) )
    {
        push( @alltime,
            $self->{status}->val( 'lastsuccess', 'end' ) +
              $self->val('waitaftersuccess') * 60 );
    }

    my ($t) = sort { $b <=> $a } @alltime;

    $t;
}

=head2 last_duration

Return duration last processing need

=cut

sub last_duration {
    my ($self) = @_;
    my $start = $self->{status}->val( 'lastrun', 'start' );
    my $end   = $self->{status}->val( 'lastrun', 'end' );
    $start && $end ? $end - $start : 0;
}

=head2 average_duration

Return the average of last time need to proccess this rsync

=cut

sub average_duration {
    my ($self) = @_;
    my @values =
      grep { defined($_) } $self->{status}->val( 'lastsuccess', 'lastduration' )
      or return 0;
    my $sum = 0;
    $sum += $_ || 0 foreach (@values);
    return $sum / scalar(@values);
}

=head2 last_start

Return last time this rsync was started

=cut

sub last_start {
    my ($self) = @_;
    $self->{status}->val( 'lastrun', 'start' ) || 0;
}

=head2 lastrun_info

Return a hash with info about last process

=cut

sub lastrun_info {
    my ($self) = @_;
    $self->{status}->ReadConfig() if ( $self->{status}->GetFileName() );
    if ( !$self->{status}->SectionExists('lastrun') ) {
        return;
    }
    return (
        status => $self->{status}->val( 'lastrun', 'status', 0 ),
        start => $self->{status}->val( 'lastrun', 'start' ) || undef,
        end   => $self->{status}->val( 'lastrun', 'end' )   || undef,
        duration => (
            $self->{status}->val( 'lastrun',   'end',   0 ) -
              $self->{status}->val( 'lastrun', 'start', 0 )
        ),
    );
}

=head2 lastsuccess_info

Return a hash with info about last process which successed

=cut

sub lastsuccess_info {
    my ($self) = @_;
    $self->{status}->ReadConfig() if ( $self->{status}->GetFileName() );
    if ( !$self->{status}->SectionExists('lastsuccess') ) {
        return;
    }
    return (
        start => $self->{status}->val( 'lastsuccess', 'start' ) || undef,
        end   => $self->{status}->val( 'lastsuccess', 'end' )   || undef,
        duration => (
            $self->{status}->val( 'lastsuccess',   'end',   0 ) -
              $self->{status}->val( 'lastsuccess', 'start', 0 )
        ),
    );
}

=head2 running_info

Retrun a hash with info about current running process if any

=cut

sub running_info {
    my ($self) = @_;
    $self->{status}->ReadConfig() if ( $self->{status}->GetFileName() );
    if ( $self->is_running() ) {
        return (
            start    => $self->{status}->val( 'job', 'start', 0 ),
            duration => scalar(time) -
              $self->{status}->val( 'job', 'start', 0 ),
            pid => $self->{status}->val( 'job', 'pid' ),
        );
    }
    else {
        return;
    }
}

=head2 choose_my_level

Find best level this mirror can expect by looking availlable mirrors
in the list

=cut

sub choose_my_level {
    my ($self) = @_;
    my ( $count, $last, $level ) = ( 0, 0, undef );
    foreach my $l ( $self->{mirrorlist}->list_levels() ) {
        my $c =
          scalar( grep { $_->level == $l } $self->{mirrorlist}->mirrors() );
        if ( $last && $c / $last < 10 ) {
            return $level + 1;
        }
        if ( $c > $count ) {
            ( $count, $level ) = ( $c, $l );
        }
        $last = $c;
    }
    return ( $level + 1 );
}

=head2 reset_mirror_list

Reset internal mirror list to start a new schedule

=cut

sub reset_mirror_list {
    my ($self) = @_;
    my %prefered_host =
      map { $_ => $self->{status}->val( 'job', "prefered_host_$_", undef ) }
      $self->{mirrorlist}->list_levels();
    $self->{mirrorlist}->reset(
        level         => $self->level,
        frequency     => $self->frequency,
        prefered_host => {%prefered_host},
    );
}

=head2 get_next_mirror

fetch the next mirror from the mirrors queue list and set it as the current
mirror to use

=cut

sub get_next_mirror {
    my ($self) = @_;
    $self->{mirrorlist}->get_next();
}

=head2 should_use_other_mirrors

Return true if others mirror than the first in the queue should
be used to success the process if first failed

=cut

sub should_use_other_mirrors {
    my ($self) = @_;
    my $maxfailureduration = $self->val('failure_time') || $self->frequency * 3;
    return (
        $self->{status}->val( 'lastsuccess', 'end', 0 ) + $maxfailureduration *
          60 < scalar(time) );
}

=head2 build_rsync_cmd

Return an array with all rsync arg to exec

=cut

sub build_rsync_cmd {
    my ( $self, $mirror ) = @_;
    $mirror ||= $self->{mirrorlist}->get_current();
    my @command = ( 'rsync', '-aqH' );

    my %woarg = (
        'delete-after'    => undef,
        'delete-excluded' => undef,
        'delete'          => undef,
    );

    foreach my $arg ( keys %woarg ) {
        if ( $self->val($arg) && lc( $self->val($arg) ) =~ /yes|true|on|1/ ) {
            push( @command, $woarg{$arg} || "--$arg", );
        }
    }

    my %warg = (
        tempdir    => '--temp-dir',
        partialdir => '--partial-dir',
        bwlimit    => undef,
    );

    foreach my $arg ( keys %warg ) {
        if ( $self->val($arg) ) {
            push( @command, $warg{$arg} || "--$arg", $self->val($arg), );
        }
    }

    if ( $self->val('exclude') ) {
        foreach ( map { split( m/ /, $_ ) } $self->val('exclude') ) {
            push( @command, '--exclude', $_ );
        }
    }

    my $url = $mirror->{url};
    if ( $mirror->{ssh} ) {
        push( @command, '-e', $self->val( 'ssh_path', 'ssh' ) );
        $url = sprintf( '%s%s:%s',
            $self->val('ssh_user') ? $self->val('ssh_user') . '@' : '',
            $mirror->{uri}->host(),
            $mirror->{uri}->path(),
        );
    }

    if ( $self->val('rsync_opts') ) {
        push( @command, map { split( m/ /, $_ ) } $self->val('rsync_opts') );
    }

    push( @command,
        $url . ( $self->val('subdir') ? "/" . $self->val('subdir') : "" ),
        $self->{dest} );

    return @command;
}

=head2 exec_rsync

Launch rsync process, return 0 on success, 1 if any error, 2 if the
mirror cannot be used, 3 on uncoverable error

=cut

sub exec_rsync {
    my ( $self, $mirror ) = @_;
    $mirror ||= $self->{mirrorlist}->get_current();

    my @command = $self->build_rsync_cmd($mirror);

    # print join(' ', @command) . "\n";

    $ENV{RSYNC_PASSWORD} = $mirror->{password} || '-';

    my $exitstatus = 0;
    @{ $self->{rsync_err_output} } = ();

    if ( my $pid = open3( my $in, my $out, my $err, @command ) ) {
        close($in);
        my $ios = IO::Select->new( $out, $err );
        while ( my @handle = $ios->can_read() ) {
            foreach my $h (@handle) {
                my $l = <$h>;
                if ($l) {
                    chomp($l);
                    push( @{ $self->{rsync_err_output} }, $l );
                }
                $ios->remove($h) if ( eof($h) );
            }
        }

        waitpid( $pid, 0 );
        $exitstatus = $?;
    }
    else {
        $exitstatus = 0;
    }

    $self->{status}->newval( 'lastrun', 'exitcode', $? );

    # checking if we have a real failure

    if ($exitstatus) {

        # Handle system exit code
        # if (grep { ($? & 127) == $_ } ()) {
        # }

        # Rsync exit code - no way to retry
        if (
            grep { ( $? >> 8 ) == $_ } (
                1,    # Syntax or usage error
                2,    # Protocol incompatibility
            )
          )
        {
            return 2;
        }

        # This is not a failure, but normal state
        if (
            grep { ( $? >> 8 ) == $_ } (
                25,    # The --max-delete limit stopped deletions
            )
          )
        {
            return 0;
        }
        if (
            grep { ( $? >> 8 ) == $_ } (
                20,    # SIGUSR1 ou SIGINT reu
            )
          )
        {
            return 3;
        }
        return 1;
    }
    else {
        return 0;
    }

}

=head2 lock

Lock this queue job

=cut

sub lock {
    my ($self) = @_;
    $self->{status}->newval( 'job', 'pid',     $$ );
    $self->{status}->newval( 'job', 'destdir', $self->{dest} );
    $self->{status}->WriteConfig( $self->{statusfile} );
}

=head2 lock

Unlock this queue job

=cut

sub unlock {
    my ($self) = @_;
    $self->{status}->delval( 'job', 'pid' );
    $self->{status}->WriteConfig( $self->{statusfile} );
}

=head2 validate_mirror

Return true if the mirror can be use according its know info and queue status

=cut

sub validate_mirror {
    my ( $self, $mirror ) = @_;
    $mirror ||= $self->{mirrorlist}->get_current();
    $mirror or return;
    if ( $self->{mirrorlist}->fetched == 2
        && !$self->should_use_other_mirrors )
    {
        return;
    }
    if ( $mirror->level >= $self->val( 'level', $self->level + 1 ) ) {
        return;
    }
    1;
}

=cut run


Run completetly the rsync process

=cut

sub run {
    my ($self) = @_;

    $self->{status}->ReadConfig() if ( $self->{status}->GetFileName() );
    $self->is_running() and return;
    $self->is_disable   and return;

    # print "Running... $self->{dest}\n";
    if ( !-d $self->{dest} ) {
        $self->log( 'ERROR', "%s does not exists, aborting", $self->{dest} );
        return 0;
    }

    if ( !$self->lock() ) {
        $self->log( 'ERROR', "cannot lock for %s", $self->{dest} );
        return 0;
    }

    my $success = -1;
    $self->reset_mirror_list();
    while ($success) {
        my $mirror = $self->get_next_mirror() or last;
        $self->validate_mirror() or last;

        foreach ( 1 .. $self->val( 'max_try', 2 ) ) {

            $self->{status}->newval( 'job', 'start', scalar( time() ) );
            $self->{status}->SetParameterComment( 'job', 'start',
                scalar( gmtime( $self->{status}->val( 'job', 'start' ) ) ) );
            $self->{status}->WriteConfig( $self->{statusfile} );

            $success = $self->exec_rsync($mirror);

            if ( !$success ) {
                $self->{status}->newval( 'lastsuccess', 'url', $mirror->{url} );
                $self->{status}->newval( 'lastsuccess', 'host', $mirror->host );
                $self->{status}
                  ->newval( 'lastsuccess', 'source_level', $mirror->level );
                my $ph = 'prefered_host_' . $mirror->level();
                if ( !$self->{status}->val( 'job', $ph ) ) {
                    $self->{status}->newval( 'job', $ph, $mirror->host );
                }
                last;
            }
            elsif ( ( $success || 0 ) == 2 ) {

                # hard failure, don't retry
                last;
            }
            sleep(10);
        }
        if ( ( $success || 0 ) == 3 ) {
            last;
        }
    }

    if ( !$success ) {
        $self->log( 'INFO', "%s done", $self->dest() );
    }
    else {
        $self->log( 'ERROR', "Failed to process %s", $self->dest() );
        foreach ( @{ $self->{rsync_err_output} || [] } ) {
            $self->log( 'WARNING', '%s::%s %s', $self->source,
                $self->{subdir} || '', $_ );
        }
    }

    foreach my $section ( 'lastrun', ( ( !$success ) ? qw(lastsuccess) : () ) )
    {
        $self->{status}
          ->newval( $section, 'start', $self->{status}->val( 'job', 'start' ) );
        $self->{status}->SetParameterComment( $section, 'start',
            scalar( gmtime( $self->{status}->val( 'job', 'start' ) ) ) );
        $self->{status}->newval( $section, 'end', scalar( time() ) );
        $self->{status}->SetParameterComment( $section, 'end',
            scalar( gmtime( $self->{status}->val( $section, 'end' ) ) ) );
        $self->{status}->newval(
            $section,
            'lastduration',
            grep { defined($_) }
              ( $self->{status}->val( $section, 'lastduration' ) )[ -9 .. -1 ],
            $self->{status}->val( $section,   'end' ) -
              $self->{status}->val( $section, 'start' )
        );
    }
    $self->{status}->newval( 'lastrun', 'status', $success );
    $self->unlock();
    $self->{status}->WriteConfig( $self->{statusfile} );

    !$success;
}

1;

=head1 SEE ALSO

L<MMM>
L<MMM::MirrorQueue>

=head1 AUTHOR

Olivier Thauvin <nanardon@nanardon.zarb.org>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2006 Olivier Thauvin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

=cut

